/**
 * @file
 * Klassenimplementierung fuer das OpenGL-Widget
 *
 * @author
 * Philip Lehmann-Boehm (minf8180)
 */

#include <GL/gl.h>
#include <GL/glu.h>

#include "glm.h"

#include <QtGui>
#include <QtOpenGL>

#include "GLwidget.h"

#include "UnitForms.h"

#include "types.h"

int extensionSupported(const char *extension)
{
  static const GLubyte *extensions = NULL;
  const GLubyte *start;
  GLubyte *where, *terminator;

  where = (GLubyte *) strchr(extension, ' ');
  if (where || *extension == '\0')
    return 0;

  if (!extensions)
    extensions = glGetString(GL_EXTENSIONS);

  start = extensions;
  for (;;) {
    where = (GLubyte *) strstr((const char *) start, extension);
    if (!where)
      break;
    terminator = where + strlen(extension);
    if (where == start || *(where - 1) == ' ') {
      if (*terminator == ' ' || *terminator == '\0') {
        return 1;
      }
    }
    start = terminator;
  }
  return 0;
}

/**
 * Konstruktor
 * @param logic Die zu dem OpenGL-Widget zugehoerige Logikinstanz
 */
GLWidget::GLWidget(Logic * logic) {
	light = NULL;
	radcircle = NULL;
	radgreen = NULL;
	radar = NULL;
	radartisch = NULL;
	radardisplay = NULL;
	person = NULL;
	this->logic = logic;
}

/**
 * Destruktor
 */
GLWidget::~GLWidget() {
	if (light != NULL) {
		delete light;
	}
	if (radcircle != NULL) {
		delete radcircle;
	}
	if (radgreen != NULL) {
		delete radgreen;
	}
	if (person != NULL) {
		delete person;
	}
	if(radar != NULL) {
	    glmDelete(radar);
	}
	if(radardisplay != NULL) {
	    glmDelete(radardisplay);
	}
	if(radartisch != NULL) {
	    glmDelete(radartisch);
	}
}

/**
 * Liefert dem Eigentuemer des Widgets die minimal benoetigte Groesse des Widgets.
 * @return Die minimal benoetigte Groesse.
 */
QSize GLWidget::minimumSizeHint() const {
	return QSize(50, 50);
}

/**
 * Liefert dem Eigentuemer die regulaer benoetigte Groesse des Widgets.
 * @return Die regulaer benoetigte Groesse
 */
QSize GLWidget::sizeHint() const {
	return QSize(400, 400);
}

/**
 * Initialisiert alle mit OpenGL zusammenhaengenden Sachen.
 */
void GLWidget::initializeGL() {

	// Einige Stati setzen.
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glEnable(GL_DEPTH_TEST);
    glEnable(GL_TEXTURE_2D);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);

	// Die Lichtquelle setzen.
	CGColor4f amb = {0.1f, 0.1f, 0.1f, 1.0f};
	CGColor4f diff = {1.0f, 1.0f, 1.0f, 1.0f};
	CGColor4f spec = {1.0f, 1.0f, 1.0f, 1.0f};
	CGPoint3Df pos = {-1.0f, 15.0f, 1.0f};
	light = new PositionedLight(amb, diff, spec, pos);

    // Anisotrope Texturfilterung aktivieren falls verfuegbar
    #ifdef GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT
    GLfloat largest_supported_anisotropy;

	if (extensionSupported("GL_EXT_texture_filter_anisotropic"))
    {
        glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &largest_supported_anisotropy);
        glTexParameterfv(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, &largest_supported_anisotropy);
    }
    #endif

	// Die Texturen laden und den Texturen-IDs zuweisen.
	textures[0] = bindTexture(QPixmap(QString("textures/radarr2t2.png")));
	textures[1] = bindTexture(QPixmap(QString("textures/radartischr2t.png")));
	textures[2] = bindTexture(QPixmap(QString("textures/radarscheibe.png")));
	textures[3] = bindTexture(QPixmap(QString("textures/radgreen.png")));
	textures[4] = bindTexture(QPixmap(QString("textures/radardisplay.png")));
	textures[5] = bindTexture(QPixmap(QString("textures/person.png")));

	// Die Darstellung einer Person als Displayliste ablegen.
	person = new DisplayList();
	person->beginDraw();
		glBindTexture(GL_TEXTURE_2D, textures[5]);
		glRotatef(90.0f, -1.0f, 0.0f, 0.0f);
		UnitForms::drawUnitQuad();
	person->endDraw();

	// Die Darstellung der radarscheibe als Displayliste ablegen.
	radcircle = new DisplayList();
	radcircle->beginDraw();
        glPushMatrix();
            glTranslatef(0.0f, 1.45f, -0.5f);
            glRotatef(90.0f, -1.0f, 0.0f, 0.0f);
            glRotatef(180.0f, 0.0f, 0.0f, 1.0f);
            glScalef(5.0f, 5.0f, 5.0f);
            glBindTexture(GL_TEXTURE_2D, textures[2]);
            UnitForms::drawUnitQuad();
		glPopMatrix();
	radcircle->endDraw();

	// Die Darstellung der gruenen radarscheibe als Displayliste ablegen.
	radgreen = new DisplayList();
	radgreen->beginDraw();
        glPushMatrix();
            glBindTexture(GL_TEXTURE_2D, textures[3]);
            UnitForms::drawUnitQuad();
		glPopMatrix();
	radgreen->endDraw();

	// Die Darstellung des Konusses, auf dem der Boden balanciert als Displayliste ablegen.
	/*cone = new DisplayList();
	cone->beginDraw();
		glPushMatrix();
			glRotatef(270.0f, 1.0f, 0.0f, 0.0f);
			glScalef(1.5f, 2.0f, 1.5f);
			glTranslatef(0.0f, 0.0f, -0.5f);
			glBindTexture(GL_TEXTURE_2D, textures[2]);
			UnitForms::drawUnitCone(20);
		glPopMatrix();
	cone->endDraw();*/

    glPushMatrix();
        radar = glmReadOBJ("mdl/radar2.obj");
        glmUnitize(radar);
        glmFacetNormals(radar);
        glmVertexNormals(radar, 90.0);
        //glmSpheremapTexture(radar);
    glPopMatrix();

    glPushMatrix();
        radartisch = glmReadOBJ("mdl/radartisch.obj");
        glmUnitize(radartisch);
        glmFacetNormals(radartisch);
        glmVertexNormals(radartisch, 90.0);
        //glmSpheremapTexture(radartisch);
    glPopMatrix();

    glPushMatrix();
        radardisplay = glmReadOBJ("mdl/radardisplay.obj");
        glmUnitize(radardisplay);
        glmFacetNormals(radardisplay);
        glmVertexNormals(radardisplay, 90.0);
        //glmSpheremapTexture(radardisplay);
    glPopMatrix();

	// Alle Materialien sind eh gleich hier...
	CGColor4f cMaterial = {1.0f, 1.0f, 0.0f, 1.0f};
	float matShininess[] = { 20.0f };
	float matSpecular[] = { 0.0f, 0.0f, 0.0f, 1.0f};
	glMaterialfv(GL_FRONT, GL_SHININESS, matShininess);
	glMaterialfv(GL_FRONT, GL_SPECULAR, matSpecular);
	glMaterialfv(GL_FRONT, GL_AMBIENT, cMaterial);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, cMaterial);

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

/**
 * Zeichnet die OpenGL-Szene.
 */
void GLWidget::paintGL() {

	// Die Szene neu initialisieren
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	gluLookAt(2.5, 6.0, 5.3,
			0.0, 0.0, 0.0,
			0.0, 1.0, 0.0);

	// Licht an...
	//glEnable(GL_LIGHTING);
	light->on();

	// Konus zeichnen.
	//cone->call();

	// Die Winkel, in denen der Boden gekippt wird.
	float angleX, angleZ;

	// Den Schwerpunkt der Personen ermitteln.
	int personCount = logic->getPersonCount();
	angleZ = 0;
	angleX = 0;
	for (int i = 0; i < personCount; ++i) {
		Person *curPerson;
		curPerson = logic->getPerson(i);
		angleZ += curPerson->x;
		angleX += curPerson->y;
	}

	// Anhand des Schwerpunktes die Achsen entsprechend kippen
	/*if (personCount > 0) {
		angleX = (angleX / personCount) * maxAngle - maxAngle / 2.0f;
		angleZ = (angleZ / personCount) * maxAngle - maxAngle / 2.0f;
	}
	glRotatef(angleX, 1.0f, 0.0f, 0.0f);
	glRotatef(angleZ, 0.0, 0.0f, -1.0f);*/



	// Den Boden zeichnen.
	//floor->call();

    // Tisch zeichnen
	glPushMatrix();
    {
        glTranslatef(0.0f, 0.8f, -0.32f);
        //glRotatef(30.0f, 0.0f, -1.0f, 0.0f);
        glScalef(11.0f, 11.0f, 11.0f);
        glColor4f(1.0f, 1.0f, 1.0f, 0.7f);
        glBindTexture(GL_TEXTURE_2D, textures[1]);
        glmDraw(radartisch, GLM_SMOOTH | GLM_TEXTURE);
    }
    glPopMatrix();

    // Radargehaeuse zeichnen
    glPushMatrix();
    {
        glTranslatef(0.0f, 1.1f, 0.0f);
        //glRotatef(30.0f, 0.0f, -1.0f, 0.0f);
        glScalef(3.0f, 3.0f, 3.0f);
        glColor4f(1.0f, 1.0f, 1.0f, 0.7f);
        glBindTexture(GL_TEXTURE_2D, textures[0]);
        glmDraw(radar, GLM_SMOOTH | GLM_TEXTURE);
    }
    glPopMatrix();

    // Radardisplay zeichnen
    glPushMatrix();
    {
        glTranslatef(0.01f, 1.388f, 2.505f);
        //glRotatef(30.0f, 0.0f, -1.0f, 0.0f);
        glScalef(3.01f, 3.01f, 3.01f);
        glColor4f(1.0f, 1.0f, 1.0f, 0.7f);
        glBindTexture(GL_TEXTURE_2D, textures[4]);
        glmDraw(radardisplay, GLM_SMOOTH | GLM_TEXTURE);
    }
    glPopMatrix();

    glEnable(GL_BLEND);

    // Radarscheibe zeichnen
    glPushMatrix();
    {
        glTranslatef(0.0f, 1.44f, -0.5f);
        glRotatef(logic->getRotation(), 0.0f, -1.0f, 0.0f);
        glRotatef(90.0f, -1.0f, 0.0f, 0.0f);
        glRotatef(180.0f, 0.0f, 0.0f, 1.0f);
        glScalef(5.0f, 5.0f, 5.0f);
        radgreen->call();
    }
    glPopMatrix();

    // Radarscheibe zeichnen
    glPushMatrix();
    {
        //glTranslatef(0.0f, 1.1f, 0.0f);
        //glScalef(10.0f, 10.0f, 10.0f);
        radcircle->call();
    }
    glPopMatrix();

    // Die Personen auf das Radar zeichnen.
    glPushMatrix();
        glTranslatef(0.0f, 1.451f, -0.5f);
        for (int i = 0; i < personCount; ++i) {
            Person *curPerson;
            curPerson = logic->getPerson(i);
            glPushMatrix();
                glTranslatef(curPerson->x * 3.0f - 1.5f, 0.0f, curPerson->y * 3.0f - 1.5f);
                glScalef(0.3f, 0.3f, 0.3f);
                person->call();
            glPopMatrix();
        }
    glPopMatrix();

	glDisable(GL_BLEND);

	// Licht aus...
	glDisable(GL_LIGHTING);
	light->off();
}

/**
 * Setzt die Projektionsmatrix von OpenGL neu, wenn sich die Fenstergroesse aendert.
 * @param width Die neue Fensterbreite
 * @param height Die neue Fensterhoehe
 */
void GLWidget::resizeGL(int width, int height) {
	glViewport(0, 0, width, height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0, (double)width / (double)height, 0.1, 100.0);
	glMatrixMode(GL_MODELVIEW);
}

