/**
 * @file
 * Klassenimplementierung fuer OpenGL-Lichter
 *
 * @author
 * Philip Lehmann-Boehm (minf8180)
 */

#include "Light.h"


// Initialisiert die statischen Variablen fuer die Lichtverwaltung.
int Light::nextFreeLight = 0;
GLenum Light::usedLights[8];

/**
 * Konstruktor fuer das abstrakte Licht.
 * @param cAmb Der ambiente Lichtanteil
 * @param cDiff Der diffuse Lichtanteil
 * @param cSpec Der spekulare Lichtanteil
 */
Light::Light(CGColor4f cAmb, CGColor4f cDiff, CGColor4f cSpec) {

	// OpenGL kann nur 8 Lichtquellen.
	if (nextFreeLight > maxLights - 1) {
		throw new LightError;
	}

	// Herausfinden, welches das naechste freie Licht ist und dieses verwenden.
	glLight = GL_LIGHT0;
	for (int i = 0; i < nextFreeLight; ++i) {
		if (usedLights[i] >= glLight) {
			glLight = usedLights[i] + 1;
		}
	}
	// Das Licht in die Liste der benutzten Lichter eintragen.
	usedLights[nextFreeLight] = glLight;
	nextFreeLight++;

	// Lichtfarben setzen.
	glLightfv(glLight, GL_AMBIENT, cAmb);
	glLightfv(glLight, GL_DIFFUSE, cDiff);
	glLightfv(glLight, GL_SPECULAR, cSpec);
}

/**
 * Destruktor
 */
Light::~Light() {
	// Das Licht aus der Liste der benutzten Lichter streichen
	for (int i = 0; i < nextFreeLight; i++) {
		if (glLight == usedLights[i]) {
			for (int i2 = i; i2 < nextFreeLight - 1; ++i2) {
				usedLights[i2] = usedLights[i2 + 1];
			}
			break;
		}
	}
	nextFreeLight--;
	off();
}

/**
 * Schaltet die Lichtquelle ein.
 */
void Light::on() {
	glEnable(glLight);
}

/**
 * Schaltet die Lichtquelle aus.
 */
void Light::off() {
	glDisable(glLight);
}

/**
 * Konstruktor
 * @param cAmb Der ambiente Lichtanteil
 * @param cDiff Der diffuse Lichtanteil
 * @param cSpec Der spekulare Lichtanteil
 * @param dir Die Leuchtrichtung
 */
DirectedLight::DirectedLight(CGColor4f cAmb, CGColor4f cDiff, CGColor4f cSpec, CGPoint3Df dir)
			: Light(cAmb, cDiff, cSpec) {
	this->dir[0] = dir[0];
	this->dir[1] = dir[1];
	this->dir[2] = dir[2];
	this->dir[3] = 0.0f;
}

/**
 * Schaltet die Lichtquelle ein.
 */
void DirectedLight::on() {
	glLightfv (glLight, GL_POSITION, dir);
	Light::on();
}

/**
 * Konstruktor
 * @param cAmb Der ambiente Lichtanteil
 * @param cDiff Der diffuse Lichtanteil
 * @param cSpec Der spekulare Lichtanteil
 * @param pos Die Lichtposition
 */
PositionedLight::PositionedLight(CGColor4f cAmb, CGColor4f cDiff, CGColor4f cSpec, CGPoint3Df pos)
			: Light(cAmb, cDiff, cSpec) {
	this->pos[0] = pos[0];
	this->pos[1] = pos[1];
	this->pos[2] = pos[2];
	this->pos[3] = 1.0f;
}

/**
 * Setzt die Abschwaechung der Lichtquelle
 * @param constant Der konstante Abschwaechungsfaktor
 * @param linear Der lineare Abschwaechungsfaktor
 * @param quadratic Der quadratische Abschwaechungsfaktor
 */
void PositionedLight::setAttenuation(GLfloat constant, GLfloat linear, GLfloat quadratic) {
	glLightfv(glLight, GL_CONSTANT_ATTENUATION, &constant);
	glLightfv(glLight, GL_LINEAR_ATTENUATION, &linear);
	glLightfv(glLight, GL_QUADRATIC_ATTENUATION, &quadratic);
}

/**
 * Schaltet die Lichtquelle ein.
 */
void PositionedLight::on() {
	glLightfv(glLight, GL_POSITION, pos);
	Light::on();
}

/**
 * Konstruktor
 * @param cAmb Der ambiente Lichtanteil
 * @param cDiff Der diffuse Lichtanteil
 * @param cSpec Der spekulare Lichtanteil
 * @param pos Die Lichtposition
 * @param dir Die Leuchtrichtung
 * @param exponent Die Fokussierung des Spots
 * @param cutoff Der Austrittswinkel des Spots
 */
SpotLight::SpotLight(CGColor4f cAmb, CGColor4f cDiff, CGColor4f cSpec, CGPoint3Df pos, CGPoint3Df dir, GLfloat exponent, GLfloat cutOff)
		: PositionedLight(cAmb, cDiff, cSpec, pos) {
	glLightf(glLight, GL_SPOT_CUTOFF, cutOff);
	glLightf(glLight, GL_SPOT_EXPONENT, exponent);
	this->dir[0] = dir[0];
	this->dir[1] = dir[1];
	this->dir[2] = dir[2];
}

/**
 * Schaltet die Lichtquelle ein.
 */
void SpotLight::on() {
	glLightfv(glLight, GL_SPOT_DIRECTION, dir);
	PositionedLight::on();
}
