/**
 * @file
 * Klassenimplementierung fuer die Logikklasse
 *
 * @author
 * Christian Harendt (minf2964)
 * Philip Lehmann-Boehm (minf8180)
 */

#include "Logic.h"

#include <QDebug>

/**
 * Konstruktor
 */
Logic::Logic() : QObject() {

	// Kramszeug initialisieren
	rotation = 0.0;
	timeSinceLastRequest = 0;
	blockSize = 0;
	personCount = 0;
	firstRead = true;

	// Streams initialisieren
	out = new QDataStream(&block, QIODevice::WriteOnly);
	out->setVersion(QDataStream::Qt_4_0);
    in = new QDataStream(&socket);
    in->setVersion(QDataStream::Qt_4_0);

	// Ereignis, wenn Daten vom Server ankommen, mit der Verarbeitungsroutine verknuepfen
	connect(&socket, SIGNAL(readyRead()), this, SLOT(readData()));
}

/**
 * Destruktor
 */
Logic::~Logic() {
	delete out;
	delete in;
}

/**
 * Berechnet den naechsten Schritt
 * @param interval Vergangene Zeit seit dem letzten Schritt
 * @return Die Schrittberechnung hat geklappt
 */
bool Logic::nextStep(double interval) {
	timeSinceLastRequest += interval;
	// Nur wenn die konstante Zeit vergangen ist, einen Request an den Server schicken
	if (timeSinceLastRequest > serverRequestPause) {
		if (socket.state() != QAbstractSocket::ConnectedState) {
			return false;
		}
		timeSinceLastRequest = 0;
		block[0] = cmdGetPersons;
//		*out << cmdGetPersons;
		socket.write(block);
		socket.flush();

	}
	// Rotation weitersetzen
	rotation += 80.0 * interval;
	normalizeAngle(&rotation);
	return true;
}

/**
 * Liefert den momentanen Stand der Rotation
 * @return Der momentane Stand der Rotation
 */
double Logic::getRotation() {
	return rotation;
}

/**
 * Normalisiert einen Winkel in den Grenzen 0-360 Grad
 * @param angle Der zu normalisierende Winkel
 */
void Logic::normalizeAngle(double *angle) {
	while (*angle < 0.0) {
		*angle += 360.0;
	}
	while (*angle > 360.0) {
		*angle -= 360.0;
	}
}

/**
 * Baut die Verbindung zum Server auf.
 * @param host Die Serveradresse
 * @param port Der Verbindungsport
 * @return Ob die Verbindung geklappt hat oder nicht
 */
bool Logic::connectToHost(QString host, int port) {
	socket.connectToHost(host, port);
	return socket.waitForConnected(connectTimeout);
}

/**
 * Liest die eingehenden Serverdaten aus
 */
void Logic::readData() {

	// Warten, bis der int mit der Datengroesse da ist.
	if (blockSize == 0) {
		if (socket.bytesAvailable() < (int)sizeof(quint16))
			return;
		*in >> blockSize;
	}

	// Warten, bis alle Daten geschickt wurden.
	if (socket.bytesAvailable() < blockSize) {
		return;
	}

	// Irgendwelche 2 Bytes mogeln sich dazwischen. Als int einfach ueberlesen...
	if (!firstRead) {
		*in >> blockSize;
	}
	firstRead = false;

	blockSize = 0;

	// Daten auslesen und die Personen ins Array eintragen
	quint16 count;
    *in >> count;
    personCount = count;
    quint16 curSize;
    for (int i = 0; i < personCount; ++i) {
		*in >> persons[i].x;
		*in >> persons[i].y;
		*in >> curSize;
		persons[i].size = curSize;
    }
}

/**
 * Liefert die Anzahl der Personen
 * @return Die Anzahl der Personen
 */
int Logic::getPersonCount() {
	return personCount;
}

/**
 * Liefert eine bestimmte Person
 * @param i Der Index der gewuenschten Person
 * @return Die gewuenschte Person
 */
Person* Logic::getPerson(int i) {
	return &(persons[i]);
}
